#define RUBY_VERSION "1.6.8"
#define RUBY_RELEASE_DATE "2003-10-15"
#define RUBY_VERSION_CODE 168
#define RUBY_RELEASE_CODE 20031015

#define RUBY_VERSION_MAJOR 1
#define RUBY_VERSION_MINOR 6
#define RUBY_VERSION_TEENY 8
#define RUBY_RELEASE_YEAR 2003
#define RUBY_RELEASE_MONTH 10
#define RUBY_RELEASE_DAY 15
